/*
 * -----------------------------------------------------------------------
 * (c) Thomas Pornin 2014. This software is provided 'as-is', without
 * any express or implied warranty. In no event will the author be held
 * liable for any damages arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to no restriction.
 *
 * Technical remarks and questions can be addressed to:
 * <pornin@bolet.org>
 * -----------------------------------------------------------------------
 */

#include <limits.h>
#include <string.h>

#include "makwa.h"

/*
 * 2048-bit modulus, generated on my laptop on 2014-02-20 12:56 UTC.
 * I did not save the private key (it was in RAM only) so don't ask for it.
 *
 * Note that the Known Answer Tests are generated with another, different
 * modulus, for which the private key is known and used (see makeKAT.c).
 */
static const unsigned char PHC_PUB2048[] = {
	0x55, 0x41, 0x4d, 0x30, 0x01, 0x00, 0xc0, 0x84,
	0x7e, 0xa3, 0x72, 0xa4, 0xd0, 0xdb, 0xa1, 0xa3,
	0x20, 0x48, 0x89, 0x4d, 0xc7, 0x99, 0x97, 0xa1,
	0x0b, 0x84, 0x2a, 0x9d, 0xb1, 0x5f, 0xc5, 0x61,
	0x4b, 0xe5, 0xa5, 0x73, 0xba, 0xcc, 0x72, 0xa9,
	0x88, 0x0a, 0x57, 0x98, 0xa3, 0x87, 0x53, 0x9b,
	0x7a, 0x4c, 0x1c, 0x71, 0xb6, 0xb1, 0x3a, 0x84,
	0xdb, 0xad, 0xaf, 0x9b, 0x03, 0xf7, 0x6f, 0x32,
	0x70, 0x84, 0x49, 0xc4, 0xfd, 0x27, 0xd2, 0xc4,
	0xaf, 0xc9, 0xdc, 0x46, 0xc4, 0xa6, 0xbe, 0xc5,
	0x5e, 0x3a, 0x3d, 0xb1, 0xa9, 0xa2, 0x56, 0xaf,
	0x05, 0x39, 0xed, 0x2a, 0xb4, 0x48, 0xb8, 0x53,
	0xb0, 0xc1, 0xaf, 0x20, 0x7b, 0x6e, 0xa2, 0x94,
	0x06, 0x34, 0x91, 0xfb, 0x5e, 0xb2, 0xdc, 0x95,
	0x0e, 0x8e, 0x1e, 0x87, 0x19, 0xc4, 0xe5, 0x3c,
	0x06, 0xdd, 0x3e, 0x7a, 0x36, 0x4b, 0x44, 0x65,
	0x26, 0x81, 0x7d, 0xd5, 0x37, 0x3d, 0x00, 0xd6,
	0x71, 0x67, 0x59, 0x06, 0x93, 0x4d, 0xad, 0x0f,
	0x7f, 0x6c, 0xed, 0xda, 0x65, 0xb4, 0x33, 0x68,
	0xf8, 0x3b, 0xae, 0x26, 0xda, 0xc4, 0x84, 0xf0,
	0x00, 0x31, 0x8d, 0xbb, 0x74, 0x80, 0x22, 0x5c,
	0xe6, 0x0e, 0xbf, 0x3a, 0x75, 0xec, 0xa3, 0x65,
	0x6f, 0xc5, 0xa0, 0x85, 0xf0, 0xf3, 0x4e, 0xcf,
	0xa9, 0xcb, 0x72, 0x1b, 0xdb, 0xd8, 0xea, 0x37,
	0xb1, 0xd8, 0x63, 0x42, 0x2c, 0x62, 0x8c, 0x73,
	0x38, 0x5d, 0x90, 0x65, 0x4a, 0xa1, 0xd0, 0x7b,
	0x1a, 0x59, 0xf6, 0x23, 0x42, 0x94, 0x0b, 0xb4,
	0x8f, 0xb0, 0x5b, 0x31, 0x47, 0xc9, 0x4c, 0x57,
	0xd7, 0x90, 0xae, 0xc7, 0x49, 0x93, 0x3a, 0x2a,
	0x19, 0xfe, 0xc9, 0x95, 0x45, 0x37, 0x6e, 0x87,
	0x68, 0x16, 0xeb, 0x2a, 0x76, 0xac, 0x56, 0x9d,
	0x08, 0xd8, 0xe1, 0xfe, 0x51, 0x81, 0xdf, 0xfb,
	0x97, 0x52, 0xb5, 0xfc, 0xe1, 0xe9
};

/*
 * See makwa.h.
 *
 * This function implements the API requirements for the Password Hashing
 * Competition; however, you MUST NOT use it since it uses an hard-coded
 * modulus that you cannot trust.
 */
int
PHS(void *out, size_t outlen,
	const void *in, size_t inlen,
	const void *salt, size_t saltlen,
	unsigned int t_cost, unsigned int m_cost)
{
	makwa_context *mc;
	int err;

	(void)m_cost;
#if UINT_MAX > LONG_MAX
	if ((unsigned long)t_cost > (unsigned long)LONG_MAX) {
		return MAKWA_BADPARAM;
	}
#endif
	mc = makwa_new();
	if (mc == NULL) {
		return MAKWA_NOMEM;
	}
	err = makwa_init(mc, PHC_PUB2048, sizeof PHC_PUB2048, MAKWA_SHA256);
	if (err < 0) {
		makwa_free(mc);
		return err;
	}
	err = makwa_hash(mc, in, inlen, salt, saltlen,
		1, outlen, (long)t_cost, out, NULL);
	makwa_free(mc);
	return err;
}
